library stage;

import 'dart:collection';
import 'dart:ui';

import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

typedef StageItemBuilder<Layer> = Widget Function(BuildContext, Layer layer);
typedef StageBuilder<Layer> = Size Function(
    BoxConstraints constraints, SelectStageItem<Layer> selectItem);
typedef SelectStageItem<Layer> = RenderBox Function(Layer layer);

/// Build arbitrarily complex layout with imperative code.
class Stage<Layer> extends RenderObjectWidget {
  final StageItemBuilder<Layer> itemBuilder;
  final StageBuilder<Layer> stageBuilder;
  final Listenable rebuild;

  /// [stageBuilder] Layout & setting up composing layer for items.
  /// [itemBuilder] Factory of item.
  /// [rebuild] A [Listenable] to trigger this [Stage] to rebuild.
  Stage(
      {@required this.stageBuilder,
      @required this.itemBuilder,
      this.rebuild,
      Key key})
      : super(key: key);

  @override
  StateElement<Layer> createElement() => StateElement<Layer>(this);

  @override
  RenderStage<Layer> createRenderObject(BuildContext context) =>
      RenderStage<Layer>();

  @override
  bool operator ==(other) =>
      other is Stage<Layer> &&
      other.runtimeType == runtimeType &&
      other.itemBuilder == itemBuilder &&
      other.stageBuilder == stageBuilder &&
      other.rebuild == rebuild &&
      other.key == key;

  @override
  int get hashCode =>
      runtimeType.hashCode ^
      itemBuilder.hashCode ^
      stageBuilder.hashCode ^
      rebuild.hashCode ^
      key.hashCode;
}

class StateElement<Layer> extends RenderObjectElement {
  StateElement(Stage<Layer> widget) : super(widget);

  @override
  Stage<Layer> get widget => super.widget;

  @override
  void mount(Element parent, newSlot) {
    super.mount(parent, newSlot);
    renderObject.stageBuilder = widget.stageBuilder;
    renderObject.itemBuilder = widget.rebuild;

    renderObject.beforeLayout = _beforeLayout;
    renderObject.getChild = _getOrCreateChildByLayer;
    renderObject.afterLayout = _afterLayout;

    renderObject.children = _childRenderObjects;
    renderObject.childrenInverse = _childRenderObjectInverse;
  }

  @override
  void update(Stage<Layer> newWidget) {
    super.update(newWidget);
    renderObject.stageBuilder = newWidget.stageBuilder;
    renderObject.itemBuilder = newWidget.rebuild;
    markNeedsBuild();
  }

  SplayTreeMap<Layer, Element> _oldChildren = SplayTreeMap();
  Map<Key, Element> _oldChildrenByKey = Map();

  SplayTreeMap<Layer, Element> _children = SplayTreeMap();

  Iterable<RenderBox> _childRenderObjects() sync* {
    for (final Element child in _children.values) {
      yield child.renderObject;
    }
  }

  Iterable<RenderBox> _childRenderObjectInverse() sync* {
    Layer layer = _children.lastKey();
    while (layer != null) {
      yield _children[layer].renderObject;
      layer = _children.lastKeyBefore(layer);
    }
  }

  void _beforeLayout() {
    final empty = _oldChildren;
    _oldChildren = _children;
    _children = empty;

    for (final child in _oldChildren.values) {
      if (child.widget.key != null) {
        assert(_oldChildrenByKey[child.widget.key] == null);
        _oldChildrenByKey[child.widget.key] = child;
      }
    }
  }

  Element _updateFromPossibleExisting(Element element, Widget childWidget) {
    if (!Widget.canUpdate(element?.widget, childWidget)) return null;
    _oldChildren.remove(element.slot);
    _oldChildrenByKey.remove(element.widget.key);
    return element;
  }

  RenderBox _getOrCreateChildByLayer(Layer layer) {
    assert(layer != null);
    final childWidget = widget.itemBuilder(this, layer);
    assert(childWidget != null);

    final Element maybeExisting = _children[layer] ??
        _updateFromPossibleExisting(
            _oldChildrenByKey[childWidget.key], childWidget) ??
        _updateFromPossibleExisting(_oldChildren[layer], childWidget);

    Element child;
    owner.buildScope(this, () {
      // TODO support GlobalKey
      child = updateChild(maybeExisting, childWidget, layer);
    });
    _children[layer] = child;

    return child.renderObject;
  }

  void _afterLayout() {
    for (final oldChild in _oldChildren.values) {
      updateChild(oldChild, null, null);
    }
    _oldChildren.clear();
    _oldChildrenByKey.clear();
  }

  @override
  void insertChildRenderObject(RenderObject child, dynamic slot) {
    renderObject.adoptChild(child);
  }

  @override
  void moveChildRenderObject(RenderObject child, dynamic slot) {}

  @override
  void removeChildRenderObject(RenderObject child) {
    renderObject.dropChild(child);
  }

  @override
  void forgetChild(Element child) {}

  @override
  void visitChildren(ElementVisitor visitor) {
    for (Element child in _children.values) {
      visitor(child);
    }
  }

  RenderStage<Layer> get renderObject => super.renderObject;
}

class RenderStage<Layer> extends RenderBox
    with
        RenderWithLayoutAlgorithm<Layer>,
        RenderWithMarkNeedsLayoutByListenable {
  RenderStage();
  Function() afterLayout;
  Function() beforeLayout;

  @override
  void performLayout() {
    invokeLayoutCallback((_) {
      beforeLayout();
      super.performLayout();
      afterLayout();
    });
  }

  @override
  void setupParentData(RenderObject child) {
    if (child.parentData is! StageParentData)
      child.parentData = StageParentData();
  }

  Iterable<RenderBox> Function() children = () => Iterable<RenderBox>.empty();
  Iterable<RenderBox> Function() childrenInverse =
      () => Iterable<RenderBox>.empty();

  @override
  void paint(PaintingContext context, Offset offset) {
    for (final child in children()) {
      final StageParentData parentData = child.parentData;

      parentData.composite.fold<PaintingContextCallback>(
        (context, offset) => context.paintChild(child, offset),
        (painter, layer) => (context, offset) =>
            context.pushLayer(layer(offset), painter, offset),
      )(context, offset + parentData.offset);
    }
  }

  @override
  bool hitTestChildren(BoxHitTestResult result, {Offset position}) {
    var isHit = false;
    for (final child in childrenInverse()) {
      isHit = isHit || child.hitTest(result, position: position);
      if (isHit) break;
    }
    return isHit;
  }

  @override
  void visitChildren(visitor) {
    for (final child in children()) {
      visitor(child);
    }
  }

  @override
  void visitChildrenForSemantics(visitor) {
    for (final child in childrenInverse()) {
      visitor(child);
    }
  }

  @override
  void markNeedsSemanticsUpdate() {
    // TODO: implement markNeedsSemanticsUpdate
    super.markNeedsSemanticsUpdate();
  }

  @override
  void detach() {
    super.detach();
    visitChildren((child) => child.detach());
  }

  @override
  void redepthChildren() {
    for (final child in childrenInverse()) {
      child.redepthChildren();
    }
  }

  @override
  List<DiagnosticsNode> debugDescribeChildren() {
    final List<DiagnosticsNode> value = <DiagnosticsNode>[];
    for (final child in children()) {
      value.add(child.toDiagnosticsNode());
    }
    return value;
  }
}

mixin RenderWithLayoutAlgorithm<Layer> on RenderBox {
  StageBuilder<Layer> _layoutAlgorithm;

  StageBuilder<Layer> get stageBuilder => _layoutAlgorithm;

  set stageBuilder(StageBuilder<Layer> layoutAlgorithm) {
    _layoutAlgorithm = layoutAlgorithm;
  }

  @override
  void performLayout() {
    size = stageBuilder(constraints, getChild);
  }

  SelectStageItem<Layer> _getChild;

  SelectStageItem<Layer> get getChild => _getChild;

  set getChild(SelectStageItem<Layer> children) {
    if (children == _getChild) return;
    _getChild = children;
  }
}

mixin RenderWithMarkNeedsLayoutByListenable on RenderObject {
  Listenable _relayout;

  Listenable get itemBuilder => _relayout;

  set itemBuilder(Listenable relayout) {
    if (relayout == _relayout) return;
    if (_relayout != null) _relayout.removeListener(markNeedsLayout);
    if (relayout != null) relayout.addListener(markNeedsLayout);
    _relayout = relayout;
  }

  @override
  void attach(PipelineOwner owner) {
    super.attach(owner);
    if (_relayout != null) _relayout.addListener(markNeedsLayout);
  }

  @override
  void detach() {
    if (_relayout != null) _relayout.removeListener(markNeedsLayout);
    super.detach();
  }
}

class StageParentData<Layer> extends BoxParentData {
  /// applied left-to-right
  List<ContainerLayer Function(Offset)> composite = [];
}

class SingleItemStage extends StatelessWidget {
  final Size Function(BoxConstraints constraints, RenderBox item) stageBuilder;
  final Listenable rebuild;

  final Widget child;

  SingleItemStage(
      {@required this.stageBuilder,
      @required this.child,
      this.rebuild,
      Key key})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Stage(
        itemBuilder: _builder, rebuild: rebuild, stageBuilder: _stageBuilder);
  }

  Widget _builder(BuildContext _, dynamic __) {
    return child;
  }

  Size _stageBuilder(BoxConstraints constraints, SelectStageItem getChild) {
    return stageBuilder(constraints, getChild(0));
  }

  @override
  bool operator ==(other) =>
      other is SingleItemStage &&
      other.runtimeType == runtimeType &&
      other.child == child &&
      other.stageBuilder == stageBuilder &&
      other.rebuild == rebuild &&
      other.key == key;

  @override
  int get hashCode =>
      runtimeType.hashCode ^
      child.hashCode ^
      stageBuilder.hashCode ^
      rebuild.hashCode ^
      key.hashCode;
}
